@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1997,1998 Shigeru Chiba, University of Tsukuba.
+  Copyright (C) 1997-2001 Shigeru Chiba, Tokyo Institute of Technology.
 
   Permission to use, copy, distribute and modify this software and   
   its documentation for any purpose is hereby granted without fee,        
@@ -27,6 +27,9 @@ ClassArray* Class::class_list = nil;
 int Class::num_of_cmd_options = 0;
 char* Class::cmd_options[];
 
+char* Class::metaclass_for_c_functions = nil;
+Class* Class::for_c_functions = nil;
+
 Ptree* Class::class_t = new LeafReserved("class", 5);
 Ptree* Class::empty_block_t = new PtreeClassBody(new Leaf("{", 1),
 						 nil,
@@ -48,7 +51,7 @@ void Class::Construct(Environment* e, Ptree* name)
 
     encode.SimpleName(name);
     def = Ptree::List(name, nil, empty_block_t);
-    def = new PtreeClassSpec(class_t, def, encode.Get());
+    def = new PtreeClassSpec(class_t, def, nil, encode.Get());
 
     full_definition = def;
     definition = def;
@@ -65,7 +68,7 @@ void Class::Construct(Environment* e, Ptree* name)
     SetEnvironment(new Environment(e));
 }
 
-void Class::InitializeInstance(Ptree* def, Ptree* margs)
+void Class::InitializeInstance(Ptree* def, Ptree*)
 {
     full_definition = def;
     if(def->Car()->IsLeaf())
@@ -100,6 +103,14 @@ char* Class::MetaclassName()
     return "Class";
 }
 
+Ptree* Class::Comments()
+{
+    if (definition->IsA(ntClassSpec))
+	return ((PtreeClassSpec*)definition)->GetComments();
+    else
+	return nil;
+}
+
 Ptree* Class::Name()
 {
     return definition->Second();
@@ -121,14 +132,8 @@ Class* Class::NthBaseClass(int n)
     while(bases != nil){
 	bases = bases->Cdr();		// skip : or ,
 	if(n-- == 0){
-	    // Bind* bind; // (not used)
-	    bool is_type;
-	    TypeInfo t;
 	    Ptree* base_class = bases->Car()->Last()->Car();
-	    if(class_environment->Lookup(base_class, is_type, t))
-		return t.ClassMetaobject();
-	    else
-		return nil;
+	    return class_environment->LookupClassMetaobject(base_class);
 	}
 
 	bases = bases->Cdr();
@@ -146,13 +151,10 @@ bool Class::IsSubclassOf(Ptree* name)
 	if(base_class->Eq(name))
 	    return TRUE;
 	else{
-	    bool is_type;
-	    TypeInfo t;
-	    if(class_environment->Lookup(base_class, is_type, t)){
-		Class* metaobject = t.ClassMetaobject();
-		if(metaobject != nil && metaobject->IsSubclassOf(name))
-		    return TRUE;
-	    }
+	    Class* metaobject
+		= class_environment->LookupClassMetaobject(base_class);
+	    if(metaobject != nil && metaobject->IsSubclassOf(name))
+		return TRUE;
 	}
 
 	bases = bases->Cdr();
@@ -340,7 +342,7 @@ bool Class::LookupMemberType(Ptree* name, TypeInfo& mem_type)
 
 // translation of classes
 
-void Class::TranslateClass(Environment* env)
+void Class::TranslateClass(Environment* e)
 {
 }
 
@@ -402,6 +404,7 @@ void Class::AppendBaseClass(Ptree* name, int specifier, bool is_virtual)
 	break;
     default :
 	MopErrorMessage("Class::AppendBaseClass()", "bad specifier");
+	lf = nil;
 	break;
     }
 
@@ -458,7 +461,7 @@ void Class::CheckValidity(char* name)
 // TranslateMemberFunction() is invoked only if the function is
 // implemented out of the class declaration (not inlined.)
 
-void Class::TranslateMemberFunction(Environment* env, Member& mem)
+void Class::TranslateMemberFunction(Environment*, Member& m)
 {
 }
 
@@ -475,7 +478,7 @@ ChangedMemberList::Cmem* Class::GetChangedMember(Ptree* decl)
 /*
   init is either "= <expression>" or "( <expression> )".
 */
-Ptree* Class::TranslateInitializer(Environment* env, Ptree* var_name,
+Ptree* Class::TranslateInitializer(Environment* env, Ptree*,
 				   Ptree* init)
 {
     if(init->Car()->Eq('('))
@@ -602,7 +605,7 @@ Ptree* Class::TranslateMemberRead(Environment* env, Ptree* object,
 	return new PtreeArrowMemberExpr(object, rest);
 }
 
-Ptree* Class::TranslateMemberRead(Environment* env, Ptree* member)
+Ptree* Class::TranslateMemberRead(Environment*, Ptree* member)
 {
     return member;
 }
@@ -611,8 +614,13 @@ Ptree* Class::TranslateMemberWrite(Environment* env, Ptree* object,
 				   Ptree* op, Ptree* member, Ptree* assign_op,
 				   Ptree* expr)
 {
-    Ptree* left;
+    // Note: If this function is invoked, TranslateAssign() on the
+    // member does not work.  Suppose that the expression is p->m = 3.
+    // Although TranslateMemberWrite() is invoked on p's class,
+    // TranslateAssign() is not invoked on m's class.  This is a sort
+    // of bug, but I don't know how to fix.
 
+    Ptree* left;
     object = TranslateExpression(env, object),
     left = Ptree::List(op, member);
     if(op->Eq('.'))
@@ -632,6 +640,68 @@ Ptree* Class::TranslateMemberWrite(Environment* env, Ptree* member,
 					   TranslateExpression(env, expr)));
 }
 
+Ptree* Class::TranslateUnaryOnMember(Environment* env, Ptree* unary_op,
+				     Ptree* object, Ptree* access_op,
+				     Ptree* member_name)
+{
+    Ptree* right;
+    object = TranslateExpression(env, object),
+    right = Ptree::List(access_op, member_name);
+    if(access_op->Eq('.'))
+	right = new PtreeDotMemberExpr(object, right);
+    else
+	right = new PtreeArrowMemberExpr(object, right);
+
+    return new PtreeUnaryExpr(unary_op, Ptree::List(right));
+}
+
+Ptree* Class::TranslateUnaryOnMember(Environment*, Ptree* unary_op,
+				     Ptree* member_name)
+{
+    return new PtreeUnaryExpr(unary_op, Ptree::List(member_name));
+}
+
+Ptree* Class::TranslatePostfixOnMember(Environment* env,
+				       Ptree* object, Ptree* access_op,
+				       Ptree* member_name, Ptree* postfix_op)
+{
+    Ptree* left;
+    object = TranslateExpression(env, object),
+    left = Ptree::List(access_op, member_name);
+    if(access_op->Eq('.'))
+	left = new PtreeDotMemberExpr(object, left);
+    else
+	left = new PtreeArrowMemberExpr(object, left);
+
+    return new PtreePostfixExpr(left, Ptree::List(postfix_op));
+}
+
+Ptree* Class::TranslatePostfixOnMember(Environment*,
+				       Ptree* member_name, Ptree* postfix_op)
+{
+    return new PtreePostfixExpr(member_name, Ptree::List(postfix_op));
+}
+
+Ptree* Class::TranslatePointer(Environment*, Ptree* var_name)
+{
+    return var_name;
+}
+
+Ptree* Class::TranslateUserStatement(Environment* env, Ptree*,
+				     Ptree*,
+				     Ptree* keyword, Ptree*)
+{
+    ErrorMessage(env, "unsupported user statement: ", keyword, keyword);
+    return nil;
+}
+
+Ptree* Class::TranslateStaticUserStatement(Environment* env,
+					   Ptree* keyword, Ptree*)
+{
+    ErrorMessage(env, "unsupported user statement: ", keyword, keyword);
+    return nil;
+}
+
 Ptree* Class::StripClassQualifier(Ptree* qualified_name)
 {
     if(qualified_name->IsLeaf())
@@ -640,25 +710,6 @@ Ptree* Class::StripClassQualifier(Ptree* qualified_name)
 	return Ptree::First(Ptree::Last(qualified_name));
 }
 
-Ptree* Class::TranslatePointer(Environment*, Ptree* var_name)
-{
-    return var_name;
-}
-
-Ptree* Class::TranslateUserStatement(Environment* env, Ptree* object,
-				     Ptree* op,
-				     Ptree* keyword, Ptree* rest)
-{
-    ErrorMessage(env, "unsupported user statement: ", keyword, keyword);
-    return nil;
-}
-
-Ptree* Class::TranslateStaticUserStatement(Environment* env,
-					   Ptree* keyword, Ptree* rest)
-{
-    ErrorMessage(env, "unsupported user statement: ", keyword, keyword);
-    return nil;
-}
 
 // utilities for translation
 
@@ -718,7 +769,8 @@ void Class::SetEnvironment(Environment* e)
 
 bool Class::AcceptTemplate()
 {
-    return FALSE;
+    return FALSE;	// Only the subclasses of TemplateClass can
+			// return true.
 }
 
 /*
@@ -828,6 +880,11 @@ void Class::ChangeDefaultMetaclass(char* name)
     Walker::ChangeDefaultMetaclass(name);
 }
 
+void Class::SetMetaclassForFunctions(char* name)
+{
+    metaclass_for_c_functions = name;
+}
+
 void Class::InsertBeforeStatement(Environment* env, Ptree* p)
 {
     Walker* w = env->GetWalker();
@@ -1046,6 +1103,13 @@ bool TemplateClass::AcceptTemplate()
     return TRUE;
 }
 
+/*
+  Translate a template instantiation.
+*/
+Ptree* TemplateClass::TranslateInstantiation(Environment*, Ptree* spec)
+{
+    return spec;
+}
 
 // class ClassArray
 
