@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1997,1998 Shigeru Chiba, University of Tsukuba.
+  Copyright (C) 1997-2001 Shigeru Chiba, Tokyo Institute of Technology.
 
   Permission to use, copy, distribute and modify this software and   
   its documentation for any purpose is hereby granted without fee,        
@@ -63,6 +63,14 @@ void Walker::NewScope()
     env = new Environment(env);
 }
 
+void Walker::NewScope(Environment* e)
+{
+    if (e != nil)
+	env = e;
+    else
+	env = new Environment(env);
+}
+
 void Walker::NewScope(Class* metaobject)
 {
     env = new Environment(env);
@@ -81,16 +89,12 @@ void Walker::RecordBaseclassEnv(Ptree* bases)
 {
     while(bases != nil){
 	bases = bases->Cdr();		// skip : or ,
-	bool is_type;
-	TypeInfo t;
 	Ptree* base_class = bases->Car()->Last()->Car();
-	if(env->Lookup(base_class, is_type, t)){
-	    Class* metaobject = t.ClassMetaobject();
-	    if(metaobject != nil){
-		Environment* e = metaobject->GetEnvironment();
-		if(e != nil)
-		    env->AddBaseclassEnv(e);
-	    }
+	Class* metaobject = env->LookupClassMetaobject(base_class);
+	if(metaobject != nil){
+	    Environment* e = metaobject->GetEnvironment();
+	    if(e != nil)
+		env->AddBaseclassEnv(e);
 	}
 
 	bases = bases->Cdr();
@@ -201,9 +205,8 @@ Ptree* Walker::TranslateTemplateClass(Ptree* temp_def, Ptree* class_spec)
 	return temp_def;
     else
 	return new PtreeTemplateDecl(temp_def->Car(),
-				     Ptree::ShallowSubst(class_spec2,
-							 class_spec,
-							 temp_def->Cdr()));
+				     Ptree::Subst(class_spec2, class_spec,
+						  temp_def->Cdr()));
 }
 
 Class* Walker::MakeTemplateClassMetaobject(Ptree* def, Ptree* userkey,
@@ -232,6 +235,74 @@ Ptree* Walker::TranslateTemplateFunction(Ptree* temp_def, Ptree* fun)
     return temp_def;
 }
 
+Ptree* Walker::TranslateTemplateInstantiation(Ptree *inst_spec)
+{
+    Ptree* userkey;
+    Ptree* class_spec;
+    Ptree* full_class_spec = Ptree::First(inst_spec);
+
+    if(full_class_spec->Car()->IsLeaf()){
+	userkey = nil;
+	class_spec = full_class_spec;
+    }
+    else{
+	userkey = full_class_spec->Car();
+	class_spec = full_class_spec->Cdr();
+    }
+
+    Class* metaobject = nil;
+    metaobject = MakeTemplateInstantiationMetaobject(full_class_spec,
+						     userkey, class_spec);
+    return TranslateTemplateInstantiation(inst_spec, userkey, 
+					  class_spec, metaobject);
+}
+
+Class* Walker::MakeTemplateInstantiationMetaobject(
+    Ptree* full_class_spec, Ptree* userkey, Ptree* class_spec)
+{
+    // [class [foo [< ... >]]] -> [class foo]
+    Ptree* class_name = Ptree::First(Ptree::Second(class_spec));
+    Bind* binding = nil;
+    if (!env->Lookup(class_name,binding))
+	return nil;
+
+    Class* metaobject = nil;
+    if (binding->What() != Bind::isTemplateClass) {
+	ErrorMessage("not declarated as a template class?!?",
+		     nil, full_class_spec);
+	metaobject = nil;
+    }
+    else
+	metaobject = binding->ClassMetaobject();
+
+    if (metaobject == nil)
+	metaobject = new TemplateClass;
+    else
+	if(metaobject->AcceptTemplate())
+	    return metaobject;
+	else{
+	    ErrorMessage("the specified metaclass is not for templates.",
+			 nil, full_class_spec);
+	    metaobject = new TemplateClass;
+	}
+
+    return metaobject;
+}
+
+Ptree* Walker::TranslateTemplateInstantiation(Ptree* inst_spec,
+		Ptree* userkey, Ptree* class_spec, Class* metaobject)
+{
+    if (metaobject == nil)
+	return inst_spec;
+    else {
+	Ptree *class_spec2 = TranslateClassSpec(class_spec);
+	if (class_spec == class_spec2)
+	    return inst_spec;
+	else
+	    return class_spec2;
+    }
+}
+
 Ptree* Walker::TranslateMetaclassDecl(Ptree* decl)
 {
     env->RecordMetaclassName(decl);
@@ -250,6 +321,36 @@ Ptree* Walker::TranslateLinkageSpec(Ptree* def)
 						body2));
 }
 
+Ptree* Walker::TranslateNamespaceSpec(Ptree* def)
+{
+    Ptree* name = Ptree::Second(def);
+    Ptree* body = Ptree::Third(def);
+
+    Environment* name_space = env->LookupNamespace0(name);
+    NewScope(name_space);
+    Ptree* body2 = Translate(body);
+    Environment* name_space2 = ExitScope();
+
+    if (name_space == nil)	// if this is the first time
+	env->RecordNamespace(name_space2, name);
+
+    if(body == body2)
+	return def;
+    else
+	return new PtreeNamespaceSpec(Ptree::First(def),
+				      Ptree::List(name, body2));
+}
+
+Ptree* Walker::TranslateUsing(Ptree* def)
+{
+    if (((PtreeUsing*)def)->isNamespace())
+	env->RecordUsingNamespace(def);
+    else
+	env->RecordUsing(def);
+
+    return def;
+}
+
 Ptree* Walker::TranslateDeclaration(Ptree* def)
 {
     Ptree* decls = Ptree::Third(def);
@@ -324,7 +425,7 @@ Ptree* Walker::TranslateDeclarators(Ptree* decls, bool record)
 	    q = TranslateDeclarator(FALSE, (PtreeDeclarator*)p);
 	    if(exp != exp2){
 		// exp2 should be a list, but...
-		if(exp2->IsLeaf())
+		if(exp2 != nil && exp2->IsLeaf())
 		    exp2 = Ptree::List(exp2);
 
 		if(p == q){
@@ -403,7 +504,7 @@ bool Walker::GetArgDeclList(PtreeDeclarator* decl, Ptree*& args)
     return FALSE;
 }
 
-Ptree* Walker::TranslateArgDeclList(bool record, Ptree* decl, Ptree* args)
+Ptree* Walker::TranslateArgDeclList(bool record, Ptree*, Ptree* args)
 {
     return TranslateArgDeclList2(record, env, FALSE, FALSE, 0, args);
 }
@@ -485,7 +586,7 @@ Ptree* Walker::FillArgumentName(Ptree* arg, Ptree* d, int arg_name)
     }
 }
 
-Ptree* Walker::TranslateAssignInitializer(PtreeDeclarator* decl, Ptree* init)
+Ptree* Walker::TranslateAssignInitializer(PtreeDeclarator*, Ptree* init)
 {
     Ptree* exp = init->Second();
     Ptree* exp2 = Translate(exp);
@@ -495,7 +596,7 @@ Ptree* Walker::TranslateAssignInitializer(PtreeDeclarator* decl, Ptree* init)
 	return Ptree::List(init->Car(), exp2);
 }
 
-Ptree* Walker::TranslateInitializeArgs(PtreeDeclarator* decl, Ptree* init)
+Ptree* Walker::TranslateInitializeArgs(PtreeDeclarator*, Ptree* init)
 {
     return TranslateArguments(init);
 }
@@ -513,7 +614,7 @@ Ptree* Walker::TranslateFunctionImplementation(Ptree* impl)
     Ptree* tspec2 = TranslateTypespecifier(tspec);
     Environment* fenv = env->RecordDeclarator(decl);
     if(fenv == nil){
-	// shouldn't reach here.
+	// reach here if resolving the qualified name fails. error?
 	NewScope();
 	decl2 = TranslateDeclarator(TRUE, (PtreeDeclarator*)decl);
 	body2 = Translate(body);
@@ -664,7 +765,7 @@ Class* Walker::MakeClassMetaobject(Ptree* def, Ptree* userkey,
 	metaobject = opcxx_ListOfMetaclass::New(default_metaclass, class_def,
 						nil);
 	if(metaobject == nil)
-	    MopErrorMessage2("the default metaclass is not loaded: ",
+	    MopErrorMessage2("the default metaclass cannot be loaded: ",
 			     default_metaclass);
     }
 
@@ -725,7 +826,7 @@ Class* Walker::LookupMetaclass(Ptree* def, Ptree* userkey, Ptree* class_def,
 	if(mclass == nil)
 	    ErrorMessage("invalid keyword: ", userkey, class_def);
 	else{
-	    metaobject = opcxx_ListOfMetaclass::New(mclass, def,
+	    metaobject = opcxx_ListOfMetaclass::New(mclass, class_def,
 						    userkey->Third());
 	    if(metaobject == nil)
 		ErrorMessage("the metaclass associated with the"
@@ -744,12 +845,11 @@ Class* Walker::LookupBaseMetaclass(Ptree* def, Ptree* class_def,
     Class* metaobject = nil;
     Ptree* bases = class_def->Third();
     while(bases != nil){
-	Bind* bind;
 	bases = bases->Cdr();
 	Ptree* base = bases->Car()->Last()->Car();
 	bases = bases->Cdr();
-	if(env->Lookup(base, bind)){
-	    Class* m = bind->ClassMetaobject();
+	Class* m = env->LookupClassMetaobject(base);
+	if(m != nil){
 	    if(metaobject == nil)
 		metaobject = m;
 	    else if(m == nil || strcmp(metaobject->MetaclassName(),
@@ -772,7 +872,7 @@ Class* Walker::LookupBaseMetaclass(Ptree* def, Ptree* class_def,
 	return nil;
 }
 
-Ptree* Walker::TranslateClassSpec(Ptree* spec, Ptree* userkey,
+Ptree* Walker::TranslateClassSpec(Ptree* spec, Ptree*,
 				  Ptree* class_def, Class* metaobject)
 {
     if(metaobject == nil)
@@ -788,7 +888,7 @@ Ptree* Walker::TranslateClassSpec(Ptree* spec, Ptree* userkey,
 	    return new PtreeClassSpec(spec->Car(),
 				      Ptree::ShallowSubst(body2, body,
 							  spec->Cdr()),
-				      spec->GetEncodedName());
+				      nil, spec->GetEncodedName());
     }
 }
 
@@ -1186,7 +1286,7 @@ Ptree* Walker::TranslateThrow(Ptree* exp)
 	return new PtreeThrowExpr(exp->First(), Ptree::List(oprnd2));
 }
 
-void Walker::TypeofThrow(Ptree* exp, TypeInfo& t)
+void Walker::TypeofThrow(Ptree*, TypeInfo& t)
 {
     t.SetVoid();
 }
@@ -1205,7 +1305,7 @@ Ptree* Walker::TranslateSizeof(Ptree* exp)
 				   Ptree::ShallowSubst(e2, e, exp->Cdr()));
 }
 
-void Walker::TypeofSizeof(Ptree* exp, TypeInfo& t)
+void Walker::TypeofSizeof(Ptree*, TypeInfo& t)
 {
     t.SetInt();
 }
@@ -1236,8 +1336,8 @@ Ptree* Walker::TranslateNew(Ptree* exp)
     return TranslateNew2(exp, userkey, scope, op, placement, type, init);
 }
 
-Ptree* Walker::TranslateNew2(Ptree* exp, Ptree* userkey, Ptree* scope,
-			     Ptree* op, Ptree* placement,
+Ptree* Walker::TranslateNew2(Ptree* exp, Ptree*, Ptree*,
+			     Ptree*, Ptree* placement,
 			     Ptree* type, Ptree* init)
 {
     Ptree* placement2 = TranslateArguments(placement);
@@ -1301,7 +1401,7 @@ Ptree* Walker::TranslateDelete(Ptree* exp)
 						       exp->Cdr()));
 }
 
-void Walker::TypeofDelete(Ptree* exp, TypeInfo& t)
+void Walker::TypeofDelete(Ptree*, TypeInfo& t)
 {
     t.SetVoid();
 }
@@ -1311,7 +1411,7 @@ Ptree* Walker::TranslateThis(Ptree* exp)
     return exp;
 }
 
-void Walker::TypeofThis(Ptree* exp, TypeInfo& t)
+void Walker::TypeofThis(Ptree*, TypeInfo& t)
 {
     t.Set(env->LookupThis());
 }
@@ -1416,7 +1516,7 @@ Ptree* Walker::TranslateUserStatement(Ptree* exp)
     return exp;
 }
 
-void Walker::TypeofUserStatement(Ptree* exp, TypeInfo& t)
+void Walker::TypeofUserStatement(Ptree*, TypeInfo& t)
 {
     t.Unknown();
 }
@@ -1474,14 +1574,14 @@ Ptree* Walker::TranslateStaticUserStatement(Ptree* exp)
     return exp;
 }
 
-void Walker::TypeofStaticUserStatement(Ptree* exp, TypeInfo& t)
+void Walker::TypeofStaticUserStatement(Ptree*, TypeInfo& t)
 {
     t.Unknown();
 }
 
 Ptree* Walker::TranslateNewDeclarator(Ptree* decl)
 {
-    Ptree* decl2;
+    Ptree* decl2 = decl;
     Ptree* p = decl;
     while(p != nil){
 	Ptree* head = p->Car();
@@ -1564,6 +1664,23 @@ Ptree* Walker::TranslateArguments(Ptree* arglist)
 	return arglist;
 }
 
+void Walker::SetDeclaratorComments(Ptree* def, Ptree* comments)
+{
+    if (def == nil || !def->IsA(ntDeclaration))
+	return;
+
+    Ptree* decl;
+    int n = 0;
+    for (;;) {
+	int i = n++;
+	decl = NthDeclarator(def, i);
+	if (decl == nil)
+	    break;
+	else if (decl->IsA(ntDeclarator))
+	    ((PtreeDeclarator*)decl)->SetComments(comments);
+    }
+}
+
 Ptree* Walker::NthDeclarator(Ptree* def, int& nth)
 {
     Ptree* decls = def->Third();
@@ -1575,7 +1692,7 @@ Ptree* Walker::NthDeclarator(Ptree* def, int& nth)
 	    return decls;
     }
     else
-	while(decls != nil){
+	while(decls != nil && !decls->IsLeaf()){
 	    if(nth-- == 0)
 		return decls->Car();
 
